import abc
from pathlib import Path
from typing import Any

from volsegtools.abc import DataHandle


class Converter(abc.ABC):
    """Converts the contents of some file format into the internal data
    structure that is then going to be used further in the processing.
    """

    @staticmethod
    @abc.abstractmethod
    async def transform_volume(input_path: Path) -> DataHandle:
        """Transforms volumetric data into a zarr array.

        Parameters
        ----------
        input_path: Path
            Path to the transformation target.

        Returns
        -------
        Lazy reference to the binary blob data.
        """
        ...

    @staticmethod
    @abc.abstractmethod
    async def transform_segmentation(input_path: Path) -> DataHandle:
        """Transforms the segmentation data into a zarr array.

        Parameters
        ----------
        input_path: Path
            Path to the transformation target.
        """
        ...

    @staticmethod
    @abc.abstractmethod
    async def collect_metadata(input_path) -> Any:
        """Collects metadata from a file.

        Parameters
        ----------
        input_path: Path
            Path to the transformation target.
        internal_data: Data
            Reference to the internal representation of the output. It is
            going to be changed by this method.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    async def collect_annotations(input_path) -> Any:
        """Collects annotations from a file.

        Parameters
        ----------
        input_path: Path
            Path to the transformation target.
        internal_data: Data
            Reference to the internal representation of the output. It is
            going to be changed by this method.
        """
        pass



from pathlib import Path
from typing import List

from typing_extensions import Self

from volsegtools.abc import Converter, Downsampler
from volsegtools.preprocessor import Preprocessor


class PreprocessorBuilder:
    """Allows fine-grained specification of the preprocessor."""

    def __init__(self) -> None:
        self._work_dir: Path | None = None
        self._output_dir: Path | None = None
        self._volume_sources: List[Path] = []
        self._segmentation_sources: List[Path] = []
        self._annotations_sources: List[Path] = []
        self._metadata_sources: List[Path] = []
        self._converter: Converter | None = None
        self._downsampler: Downsampler | None = None
        # Currently, we only support a single preprocessor, after that we
        # could add support for generic preprocessor
        # self._preprocessor_type: Optional[Preprocessor] = None

    def set_converter(self, converter: Converter) -> Self:
        """Sets a converter that is going to be used by the resulting
        preprocessor.

        If this method is going to be called multimple times, it is going
        to override the previously set converter.
        """
        self._converter = converter
        return self

    def set_downsampler(self, downsampler: Downsampler) -> Self:
        """Sets a downsampler that is going to be used by the resulting
        preprocessor.

        If this method is going to be called multimple times, it is going
        to override the previously set downsampler.
        """
        self._downsampler = downsampler
        return self

    def add_volume_src_file(self, file_path: Path) -> Self:
        """Adds the source file for volumetric data.

        It should be used in cases where there are multiple files and
        set_input_file cannot be used.
        """
        self._volume_sources.append(file_path)
        return self

    def add_segmentation_src_file(self, file_path: Path) -> Self:
        """Sets the source file for annotations.

        It should be used in cases where there are multiple files and
        set_input_file cannot be used.
        """
        self._segmentation_sources.append(file_path)
        return self

    def add_metadata_src_file(self, file_path: Path) -> Self:
        """Adds the source file for metadata.

        It should be used in cases where there are multiple files and
        set_input_file cannot be used.
        """
        return self

    def add_annotations_src_file(self, file_path: Path) -> Self:
        """Adds the source file for annotations.

        It should be used in cases where there are multiple files and
        set_input_file cannot be used.
        """
        return self

    def set_work_dir(self, file_path: Path) -> Self:
        """Sets the working directory of the processor.

        It is going to be added to the search path for source files of the
        preprocessor. Also, the output of the preprocessor is going to be
        saved at this location.
        """
        return self

    def set_output_dir(self, file_path: Path) -> Self:
        self._output_dir = file_path
        return self

    def build(self) -> Preprocessor:
        """Builds the resulting preprocessor."""
        if self._downsampler is None:
            raise RuntimeError("Downsampler was not set")

        if self._converter is None:
            raise RuntimeError("Converter was not set")

        return Preprocessor(
            self._downsampler,
            self._converter,
            self._volume_sources,
            self._segmentation_sources,
            self._metadata_sources,
            self._annotations_sources,
            self._work_dir,
            self._output_dir,
        )



import dataclasses
from typing import List

from volsegtools.core import LatticeKind, Vector3


@dataclasses.dataclass
class DescriptiveStatistics:
    """Represents statistics that should be collected for some data set for
    it to be representable in CIF.
    """

    mean: float
    std: float
    max: float
    min: float


@dataclasses.dataclass
class ChannelMetadata:
    id: int
    statistics: DescriptiveStatistics


@dataclasses.dataclass
class TimeFrameMetadata:
    # TODO: rename to 'name'
    axis_order: Vector3 = dataclasses.field(default_factory=Vector3)
    lattice_id: str = "unknown"
    kind: LatticeKind = dataclasses.field(default=LatticeKind.VOLUME)
    id: int = -1
    axis_order: Vector3 = dataclasses.field(default_factory=Vector3)
    resolution: int = -1
    origin: Vector3 = dataclasses.field(default_factory=Vector3)
    lattice_dimensions: Vector3 = dataclasses.field(default_factory=Vector3)
    voxel_size: Vector3 = dataclasses.field(default_factory=Vector3)
    channels: List[ChannelMetadata] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class OriginalTimeFrameMetadata(TimeFrameMetadata):
    axis_order: Vector3 = dataclasses.field(default_factory=Vector3)


@dataclasses.dataclass
class Metadata:
    original_time_frame: OriginalTimeFrameMetadata = dataclasses.field(
        default_factory=OriginalTimeFrameMetadata
    )
    time_frames: List[TimeFrameMetadata] = dataclasses.field(default_factory=list)



from pathlib import Path

import dask.array as da
import mrcfile
import numpy as np

from volsegtools.abc import Converter
from volsegtools.core import LatticeKind, Vector3
from volsegtools.model import StoringParameters, TimeFrameMetadata
from volsegtools.model.opaque_data_handle import OpaqueDataHandle
from volsegtools.model.working_store import WorkingStore


class MapConverter(Converter):
    @staticmethod
    def _normalize_axis_order(data: da.Array, header: np.recarray) -> da.Array:
        """Normalizes the order of axes in the data to (x, y, z).

        Due to the fact that we use column order we have to transpose
        the array in the end.

        Parameters
        ----------
        data: da.Array
            MCR file data with any order of axes.
        header: np.recarray
            The of the file from which comes the data.

        Returns
        -------
        da.Array
            Array with normalized order of axes. It is the view to the input
            array.
        """
        CORRECT_ORDER = (0, 1, 2)

        current_order = tuple(
            int(axis) - 1 for axis in [header.mapc, header.mapr, header.maps]
        )

        if tuple(current_order) != CORRECT_ORDER:
            da.moveaxis(data, current_order, CORRECT_ORDER)

        data.transpose()

        return data

    @staticmethod
    async def transform_volume(input_path: Path) -> OpaqueDataHandle:
        with mrcfile.mmap(input_path, "r+") as mrc:
            if mrc.data is None or mrc.header is None:
                raise RuntimeError("Failed to read data from MAP file")

            array = da.from_array(mrc.data)
            array = MapConverter._normalize_axis_order(array, mrc.header)
            internal_data = WorkingStore.instance
            internal_data.volume_dtype = mrc.data.dtype
            internal_data.is_volume_dtype_set = True
            volume_id: str = input_path.stem
            return internal_data.store_lattice_time_frame(
                StoringParameters(), array, volume_id
            )

    @staticmethod
    async def transform_segmentation(input_path: Path) -> OpaqueDataHandle:
        with mrcfile.open(input_path, "r+") as mrc:
            if mrc.data is None or mrc.header is None:
                raise RuntimeError("Failed to read data from MAP file")

            data = da.from_array(mrc.data)
            data = MapConverter._normalize_axis_order(data, mrc.header)

            if isinstance(data.dtype, np.floating):
                data = data.astype(np.byte)

            internal_data = WorkingStore.instance

            internal_data.volume_dtype = data.dtype
            internal_data.is_volume_dtype_set = True

            segmentation_id: str = input_path.stem

            storing_params = StoringParameters()
            storing_params.storage_dtype = data.dtype
            storing_params.lattice_kind = LatticeKind.SEGMENTATION
            return internal_data.store_lattice_time_frame(
                storing_params, data, segmentation_id
            )

    @staticmethod
    async def collect_annotations(input_path) -> None:
        pass

    @staticmethod
    async def collect_metadata(input_path) -> TimeFrameMetadata:
        with mrcfile.open(input_path, "r+") as mrc:
            if mrc.data is None or mrc.header is None:
                raise RuntimeError("Failed to read data from MAP file")
            lattice_shape = Vector3(
                int(mrc.header.nx),
                int(mrc.header.ny),
                int(mrc.header.nz),
            )
            header = mrc.header

        axis_order_map = {
            header.mapc - 1: 0,
            header.mapr - 1: 1,
            header.maps - 1: 2,
        }

        axis_order = Vector3(0, 1, 2)

        start = (header.nxstart, header.nystart, header.nzstart)
        start = Vector3(
            start[axis_order_map[0]],
            start[axis_order_map[1]],
            start[axis_order_map[2]],
        )

        original_voxel_size = Vector3(
            header.cella.x,
            header.cella.y,
            header.cella.z,
        )

        origin = Vector3(
            start.x * original_voxel_size.x,
            start.y * original_voxel_size.y,
            start.z * original_voxel_size.z,
        )

        original_time_frame_metadata = TimeFrameMetadata(
            axis_order=axis_order,
            lattice_id=input_path.stem,
            id=0,
            resolution=0,
            origin=origin,
            lattice_dimensions=lattice_shape,
            voxel_size=original_voxel_size,
            # As we are not converting the first resolution to BCIF, this can
            # be left empty.
            channels=[],
        )

        print("Original Metadata:", original_time_frame_metadata)

        return original_time_frame_metadata



import asyncio
from pathlib import Path
from typing import List, Optional

from volsegtools.abc import Converter, Downsampler
from volsegtools.core.lattice_kind import LatticeKind
from volsegtools.model.working_store import WorkingStore
from volsegtools.serialization import BCIFSerializer


class Preprocessor:
    state = {"downsampling_status": 0.0}

    def __init__(
        self,
        downsampler: Downsampler,
        converter: Converter,
        volume_input_files: List[Path],
        segmentation_input_files: List[Path],
        metadata_input_files: Optional[List[Path]] = None,
        annotations_input_files: Optional[List[Path]] = None,
        work_dir: Optional[Path] = None,
        output_dir: Optional[Path] = None,
    ):
        self.converter: Converter = converter
        self._output_dir = output_dir if output_dir is not None else Path()
        self._data = WorkingStore.instance
        self.downsampler = downsampler
        self._volume_input_files = volume_input_files
        self._segmentation_input_files = segmentation_input_files
        self._metadata_input_files = metadata_input_files
        self._annotations_input_files = annotations_input_files
        self._work_dir = work_dir

    async def transform_volume(self):
        raise NotImplementedError()

    async def collect_metadata(self):
        raise NotImplementedError()

    def create_converter_from_format(self):
        raise NotImplementedError()

    async def transform_segmentation(self):
        raise NotImplementedError()

    async def preprocess(self):
        raise NotImplementedError()

    def sync_preprocess(self):
        data = []

        # 1. Data conversion phase
        for path in self._volume_input_files:
            data.append(asyncio.run(self.converter.transform_volume(path)))
            data[-1].metadata = asyncio.run(self.converter.collect_metadata(path))
            # FIX: This should be automatic!
            data[-1].metadata.kind = LatticeKind.VOLUME
        for path in self._segmentation_input_files:
            data.append(asyncio.run(self.converter.transform_segmentation(path)))
            data[-1].metadata = asyncio.run(self.converter.collect_metadata(path))
            # FIX: This should be automatic!
            data[-1].metadata.kind = LatticeKind.SEGMENTATION

        downsampled_data = []
        for ref in data:
            # TODO: it has to be flattened!
            downsampled_data += asyncio.run(self.downsampler.downsample_lattice(ref))
        print("Downsampled Data:", downsampled_data)
        # 4. Serialization Phase
        for ref in downsampled_data:
            asyncio.run(BCIFSerializer.serialize(ref, self._output_dir))

    def downsample(self):
        pass



import abc
from typing import Any

from numpy.typing import ArrayLike


class DataHandle(abc.ABC):
    @abc.abstractmethod
    def access(self) -> ArrayLike:
        raise NotImplementedError()

    @abc.abstractmethod
    def unwrap(self, target: str) -> ArrayLike:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def metadata(self) -> Any:
        raise NotImplementedError()

    @metadata.setter
    @abc.abstractmethod
    def metadata(self, new_metadata: Any) -> None:
        raise NotImplementedError()



import dataclasses

import dask.array as da
import numpy as np
from numpy.typing import ArrayLike
from zarr.core.array import Array as ZarrArray

from volsegtools.abc.data_handle import DataHandle
from volsegtools.model.metadata import TimeFrameMetadata

# cuPy is an optional import to the volseg-tools (as CUDA may not be available
# everywhere)
try:
    import cupy as cp
except ImportError:
    # Define the array type as something inacessible
    ...


class TimeFrameIterator: ...


class ResolutionIterator: ...


class ChannelIterator: ...


@dataclasses.dataclass
class ChannelInfo:
    resolution: str
    time: str
    channel: str
    data: ZarrArray


class FlatChannelIterator:
    def __init__(self, group):
        self.group = group
        self._iter = self._group_iter()

    def _group_iter(self):
        for resolution, resolution_group in self.group.groups():
            for time, time_group in resolution_group.groups():
                for channel, channel_arr in time_group.arrays():
                    yield ChannelInfo(resolution, time, channel, channel_arr)

    def __iter__(self):
        return self

    def __next__(self) -> ChannelInfo:
        return next(self._iter)


class OpaqueDataHandle(DataHandle):
    """Wrapper around basic volseg-tools data model.

    The data held by the data model can potentially be very large and have to
    be storred in files in the files system (usually by leveraging Zarr store).

    This is a wrapper around this file that is passed between the different
    stages and holds the reference to the data stored in the file system. If
    unwrapped then it is directly loaded to the memory as an array.
    """

    def __init__(self, reference):
        if type(reference) is np.ndarray:
            if reference.base is None:
                raise RuntimeError("Data reference can be made only from a view")
        self._metadata = TimeFrameMetadata()
        self._internal_repr = reference

    @property
    def metadata(self) -> TimeFrameMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, new_metadata) -> None:
        self._metadata = new_metadata

    def access(self) -> ArrayLike:
        return self._internal_repr

    def unwrap(self, target="numpy") -> ArrayLike:
        """
        Supported unwrapping targets are: numpy, dask, and optionally cupy
        """
        match target:
            case "numpy":
                return self._repr_to_numpy_arr()
            case "dask":
                return self._repr_to_dask_arr()
            case "cupy":
                return self._repr_to_cupy_arr()
            case _:
                raise RuntimeError("Unknown unwrapping kind")

    def _repr_to_numpy_arr(self):
        if type(self._internal_repr) is np.ndarray:
            return self._internal_repr.copy()
        elif type(self._internal_repr) is ZarrArray:
            return self._internal_repr[:]

        raise RuntimeError(
            f"Unknown internal representation {type(self._internal_repr)}"
        )

    def _repr_to_dask_arr(self):
        return None

    def _repr_to_cupy_arr(self):
        return None



from .lattice_kind import LatticeKind
from .vector import Vector3
from .bounds import Bounds
from .gaussian_kernel_3D import Gaussian3DKernel
from .downsampling_parameters import DownsamplingParameters, to_bytes



from typing import Optional

import pydantic


# TODO: rather use base model...
@pydantic.dataclasses.dataclass
class Bounds:
    # TODO: might be int
    # TODO: this should be generic and accept any class that support some traits
    min: Optional[pydantic.NonNegativeFloat] = 0.0
    max: Optional[pydantic.NonNegativeFloat] = 0.0



import abc
from typing import Any, List

from volsegtools.abc import DataHandle


class Downsampler(abc.ABC):
    @property
    def parameters(self) -> Any: ...

    @abc.abstractmethod
    async def downsample_lattice(self, data: DataHandle) -> List[DataHandle]:
        """Downsamples the provided lattice.

        The lattice is changed in place and only a reference to the same data
        array is returned.

        Parameters
        ----------
        name: str
            Name of the lattice that should be previously stored in the lattice

        Returns
        -------
        """
        ...



import numpy as np

from volsegtools.abc import ConvolutionKernel


class Gaussian3DKernel(ConvolutionKernel):
    """Generate a 3D Gaussian kernel."""

    def __init__(self, size: int, sigma: float):
        ax = np.linspace(-(size // 2), size // 2, size)
        xx, yy, zz = np.meshgrid(ax, ax, ax, indexing="ij")
        kernel = np.exp(-(xx**2 + yy**2 + zz**2) / (2.0 * sigma**2))
        self._kernel_lattice = kernel / np.sum(kernel)

    def as_ndarray(self) -> np.ndarray:
        return self._kernel_lattice



import abc
from typing import Any


class Preprocessor(abc.ABC):
    """Processes"""

    @abc.abstractmethod
    async def transform_volume(self):
        pass

    @abc.abstractmethod
    async def transform_segmentation(self):
        pass

    @abc.abstractmethod
    async def collect_metadata(self):
        pass

    @abc.abstractmethod
    async def collect_annotation(self):
        pass

    @abc.abstractmethod
    async def downsample(self):
        pass

    @abc.abstractmethod
    async def get_results(self) -> Any:
        pass

    @abc.abstractmethod
    async def preprocess(self):
        pass

    @abc.abstractmethod
    def sync_preprocess(self):
        pass



import dataclasses
from pathlib import Path
from typing import Tuple

import dask.array as da
import numpy as np
import zarr
import zarr.storage

from volsegtools.core import LatticeKind
from volsegtools.model.chunking_mode import ChunkingMode
from volsegtools.model.metadata import Metadata
from volsegtools.model.opaque_data_handle import OpaqueDataHandle
from volsegtools.model.storing_parameters import StoringParameters


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

    @property
    def instance(cls):
        return cls._instances[cls]


class WorkingStore(metaclass=Singleton):
    def __init__(self, store_path: Path):
        self.data_store = zarr.storage.LocalStore(root=store_path)
        self.root_group = zarr.create_group(store=self.data_store)

        self._metadata = Metadata()

        self.volume_dtype = np.float64
        self.is_volume_dtype_set = False

        self.segmentation_dtype = np.float64
        self.is_segmentation_dtype_set = False

        self._volume_data_group = self.root_group.require_group("volume_data")
        self._segmentation_data_group = self.root_group.require_group(
            "segmentation_data"
        )

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value
        # TODO: it should return a dictionary
        # self.root_group.attrs.put(dataclasses.asdict(self._metadata))

    @property
    def volume_data_group(self):
        return self._volume_data_group

    @property
    def segmentation_data_group(self):
        return self._segmentation_data_group

    def get_data_array(
        self, lattice_id, resolution, time_frame, channel, kind=LatticeKind.VOLUME
    ):
        kind_group = self.get_data_group(kind)
        lattice_group = kind_group.require_group(lattice_id)
        resolution_group: zarr.Group = lattice_group.require_group(
            f"resolution_{resolution}"
        )
        time_frame_group: zarr.Group = resolution_group.require_group(
            f"time_frame_{time_frame}"
        )
        # FIX: this is unsafe, there should be some check!
        return list(time_frame_group.arrays())[channel][1][:]

    @staticmethod
    def _compute_chunk_size_based_on_data(
        data_shape: Tuple[int, ...],
    ) -> Tuple[int, ...]:
        chunks = tuple([int(i / 4) if i > 4 else i for i in data_shape])
        return chunks

    @staticmethod
    def _resolve_chunking_method(mode: ChunkingMode, data_shape: Tuple[int, ...]):
        match mode:
            case ChunkingMode.AUTO:
                return "auto"
            case ChunkingMode.NONE:
                return (0, 0)
            case ChunkingMode.CUSTOM:
                return Data._compute_chunk_size_based_on_data(data_shape)
            case _:
                raise RuntimeError("Unsupported chunking method!")

    def get_data_group(self, lattice_kind: LatticeKind):
        match lattice_kind:
            case LatticeKind.VOLUME:
                return self.volume_data_group
            case LatticeKind.SEGMENTATION:
                return self.segmentation_data_group
            case _:
                raise RuntimeError("Unknown lattice kind encountered.")

    def store_lattice_time_frame(
        self,
        params: StoringParameters,
        data: da.Array,
        lattice_id: str,
    ) -> OpaqueDataHandle:
        kind_group = self.get_data_group(params.lattice_kind)
        lattice_group = kind_group.require_group(lattice_id)
        resolution_group: zarr.Group = lattice_group.require_group(
            f"resolution_{params.resolution_level}"
        )
        time_frame_group: zarr.Group = resolution_group.require_group(
            f"time_frame_{params.time_frame}"
        )

        used_compressor = None
        if params.is_compression_enabled:
            used_compressor = params.compressor

        zarr_repr: zarr.Array = time_frame_group.create_array(
            name=str(params.channel),
            chunks=WorkingStore._resolve_chunking_method(
                params.chunking_mode, data.shape
            ),
            dtype=params.storage_dtype,
            compressors=[used_compressor] if used_compressor is not None else None,
            shape=data.shape,
            overwrite=True,
        )

        da.to_zarr(arr=data, url=zarr_repr, overwrite=True, compute=True)
        ref = OpaqueDataHandle(zarr_repr)
        ref.metadata.lattice_id = lattice_id
        return ref



import asyncio
import collections
import logging
import math
from typing import List, Tuple

import dask
import dask.array as da
import dask_image.ndfilters as dask_filter
import numpy as np

from volsegtools.core import (
    Bounds,
    DownsamplingParameters,
    LatticeKind,
    Vector3,
    to_bytes,
)
from volsegtools.downsampler import BaseDownsampler
from volsegtools.model import (
    ChannelMetadata,
    DescriptiveStatistics,
    FlatChannelIterator,
    OpaqueDataHandle,
    StoringParameters,
    TimeFrameMetadata,
)
from volsegtools.model.working_store import WorkingStore

MIN_GRID_SIZE = 100**1


class HierarchyDownsampler(BaseDownsampler):
    """ """

    # This value was used in the previous version of the preprocessor.
    KERNEL_PARAMETERS: Tuple[int, int, int] = (1, 4, 6)

    def __init__(self):
        _parameters = DownsamplingParameters()
        super().__init__(_parameters)

    """
    def __init__(self, params = DownsamplingParameters()):
        super().__init__(params)
    """

    async def downsample_lattice(
        self, data: OpaqueDataHandle
    ) -> List[OpaqueDataHandle]:
        # We have to get the actual data from the zarr store.
        store = WorkingStore.instance
        channel_iter = FlatChannelIterator(
            store.get_data_group(data.metadata.kind).require_group(
                data.metadata.lattice_id
            )
        )

        ret_value = []

        # Each time frame can have multiple channels, this flat channel
        # iterator iterates over each channel in each resolution for this
        # particular data lattice.
        for channel_info in channel_iter:
            if 1 in channel_info.data.shape:
                # TODO: Add some kind of message, that it does not make
                # sense, fi there is only a single dimension.
                continue

            # This is always true, becase we are reading from the zarr store.
            dask_arr = da.from_zarr(
                url=channel_info.data,
                chunks=channel_info.data.chunks,
            )

            current_level_data = dask_arr

            downsampling_steps = self._calculate_downsampling_steps_count(
                dask_arr,
            )
            logging.info(f"Downsampling steps {downsampling_steps}")

            downsampling_levels = self._calculate_downsampling_levels(
                dask_arr,
                downsampling_steps=downsampling_steps,
                # TODO: this should be changeable with some "additional_downsampler_config" which
                # would be just an arbitrary dictionary.
                factor=8,
            )
            logging.info(f"Downsampling levels {downsampling_levels}")

            for step in range(downsampling_steps):
                # TODO use step to compute the voxel size
                current_ratio = 2 ** (step + 1)
                logging.info(
                    f"Currently downsampling r{current_ratio}, {channel_info.time} and ch{channel_info.channel}"
                )
                downsampled_data = dask_filter.convolve(
                    current_level_data,
                    self.parameters.kernel.as_ndarray(),
                    mode="mirror",
                    cval=0.0,
                )
                # TODO: find out what this does, removes the neighbor?
                downsampled_data = downsampled_data[::2, ::2, ::2]

                if current_ratio not in downsampling_levels:
                    continue

                if self.parameters.acceptance_threshold != None:
                    logging.info("Using the acceptance threshold")
                    downsampled_data[
                        downsampled_data >= self.parameters.acceptance_threshold
                    ]

                if self.parameters.is_mask:
                    # TODO: this is inefficient point of usage, merge this
                    # acceptance threshold check
                    logging.info("Converting to Mask")
                    downsampled_data = da.where(
                        downsampled_data > self.parameters.acceptance_threshold, 1, 0
                    )

                stats = dask.compute(
                    da.mean(downsampled_data),
                    da.std(downsampled_data),
                    downsampled_data.max(),
                    downsampled_data.min(),
                )
                stats = DescriptiveStatistics(*stats)

                data_ref = OpaqueDataHandle(downsampled_data)
                data_ref.metadata = data.metadata

                # Only change things that are really different.
                data_ref.metadata.id = int(channel_info.time.split("_")[-1])
                data_ref.metadata.resolution = current_ratio
                data_ref.metadata.lattice_dimensions = Vector3(
                    downsampled_data.shape[0],
                    downsampled_data.shape[1],
                    downsampled_data.shape[2],
                )
                data_ref.metadata.channels.append(
                    ChannelMetadata(int(channel_info.channel), stats)
                )

                ret_value.append(data_ref)

                params = StoringParameters(
                    resolution_level=current_ratio,
                    time_frame=int(channel_info.time.split("_")[-1]),
                    channel=int(channel_info.channel),
                    storage_dtype=np.byte
                    if self.parameters.is_mask
                    else downsampled_data.dtype,
                    lattice_kind=data.metadata.kind,
                )
                WorkingStore.instance.store_lattice_time_frame(
                    params,
                    downsampled_data,
                    data.metadata.lattice_id,
                )
                current_level_data = downsampled_data
        return ret_value

    def _calculate_downsampling_steps_count(
        self,
        data: da.Array,
        downsampling_factor: int = 8,
    ) -> int:
        """Calculates the number of steps that are going to be taken during
        the downsampling of the input data.

        Parameters
        ----------
        data: da.Array
            The input data that shall be downsampled.
        downsampling_factor: int
            The factor of downsampling.

        Returns
        -------
        int:
            the number of downsampling steps.
        """

        steps_count: int = 0

        # Steps are calculated either from bounds provided as downsampling
        # parameters, if any. In that case the maximal bound has priority over
        # the minimal bound as it is the user's decision.
        # Otherwise we have to compute them manually.
        if self.parameters.downsampling_level_bounds:
            level_bounds: Bounds = self.parameters.downsampling_level_bounds
            if level_bounds.min:
                steps_count = int(math.log2(level_bounds.min))
            if level_bounds.max:
                steps_count = int(math.log2(level_bounds.max))
        else:
            input_grid_size: float = math.prod(data.shape)

            if input_grid_size <= MIN_GRID_SIZE:
                return 1

            file_size_in_bytes = data.dtype.itemsize * input_grid_size
            size_per_downsampling = file_size_in_bytes / to_bytes(
                self.parameters.size_per_level_bounds_in_mb.min
            )
            steps_count = int(math.log(size_per_downsampling, downsampling_factor))

        return steps_count

    def _calculate_downsampling_levels(
        self,
        data: da.Array,
        factor: int,
        downsampling_steps: int,
    ) -> List[int]:
        levels: List[int] = [2**x for x in range(1, downsampling_steps + 1)]

        if self.parameters.downsampling_level_bounds:
            level_bounds: Bounds = self.parameters.downsampling_level_bounds
            if level_bounds.max:
                predicate = lambda x: x <= level_bounds.max
                levels = [x for x in levels if predicate(x)]
            if level_bounds.min:
                predicate = lambda x: x >= level_bounds.min
                levels = [x for x in levels if predicate(x)]

        size_per_level: int = self.parameters.size_per_level_bounds_in_mb.max
        if size_per_level:
            # TODO: make this a parameters
            input_grid_size: float = math.prod(data.shape)
            file_size_in_bytes = data.dtype.itemsize * input_grid_size
            # TODO: this needs a better name
            n = math.ceil(
                math.log(
                    file_size_in_bytes
                    / (
                        # TODO: Make this a function or something
                        size_per_level * 1024**2
                    ),
                    factor,
                )
            )
            levels = [x for x in levels if x >= 2**n]

        if len(levels) == 0:
            raise RuntimeError(
                "No downsamplings could be saved because the max size per"
                f"channel ({size_per_level}) is too low"
            )
        return levels



from .chunking_mode import ChunkingMode

# from .downsampling_data import Data, ChannelInfo, FlatChannelIterator
from .metadata import (
    ChannelMetadata,
    DescriptiveStatistics,
    Metadata,
    OriginalTimeFrameMetadata,
    TimeFrameMetadata,
)
from .opaque_data_handle import ChannelInfo, FlatChannelIterator, OpaqueDataHandle
from .storing_parameters import StoringParameters
from .working_store import WorkingStore



from typing import Optional

import pydantic
from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from volsegtools.abc.kernel import ConvolutionKernel
from volsegtools.core import Bounds, Gaussian3DKernel


def _minimal_size_bounds_factory():
    return Bounds(5.0, None)


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class DownsamplingParameters:
    downsampling_level_bounds: Optional[Bounds] = None
    should_remove_original_resolution: bool = False
    is_mask: bool = False
    acceptance_threshold: Optional[pydantic.PositiveFloat] = None
    kernel: ConvolutionKernel = Gaussian3DKernel(5, 1.0)
    size_per_level_bounds_in_mb: Bounds = pydantic.Field(
        default_factory=_minimal_size_bounds_factory
    )


def to_bytes(megabytes: int):
    return megabytes * 10**6
